import ecdsa  # type: ignore[import]

SECP256_K1_N = int("fffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141", 16)


def _convert_der_to_rsv(signature: bytes, v_adjustment_factor: int = 0) -> dict:
    r, s = ecdsa.util.sigdecode_der(signature, ecdsa.SECP256k1.order)
    v = v_adjustment_factor
    if s > SECP256_K1_N / 2:
        s = SECP256_K1_N - s
    r = r.to_bytes(32, byteorder="big")
    s = s.to_bytes(32, byteorder="big")
    return dict(v=v, r=r, s=s)



from functools import cached_property
from typing import Any, Iterator, Optional

from ape.api.accounts import AccountAPI, AccountContainerAPI, TransactionAPI
from ape.types import AddressType, MessageSignature, SignableMessage, TransactionSignature
from eth_account._utils.legacy_transactions import serializable_unsigned_transaction_from_dict
from eth_account.messages import _hash_eip191_message, encode_defunct
from eth_pydantic_types import HexBytes
from eth_typing import Hash32
from eth_utils import keccak, to_checksum_address

from .client import kms_client
from .utils import _convert_der_to_rsv


class AwsAccountContainer(AccountContainerAPI):
    @property
    def aliases(self) -> Iterator[str]:
        return map(lambda x: x.alias, kms_client.raw_aliases)

    def __len__(self) -> int:
        return len(kms_client.raw_aliases)

    @property
    def accounts(self) -> Iterator[AccountAPI]:
        return map(
            lambda x: KmsAccount(
                key_alias=x.alias,
                key_id=x.key_id,
                key_arn=x.arn,
            ),
            kms_client.raw_aliases,
        )


class KmsAccount(AccountAPI):
    key_alias: str
    key_id: str
    key_arn: str

    @property
    def public_key(self):
        return kms_client.get_public_key(self.key_id)

    @cached_property
    def address(self) -> AddressType:
        return to_checksum_address(keccak(self.public_key[-64:])[-20:].hex().lower())

    def _sign_raw_hash(self, msghash: HexBytes | Hash32) -> Optional[bytes]:
        return kms_client.sign(self.key_id, msghash)

    def sign_raw_msghash(self, msghash: HexBytes | Hash32) -> Optional[MessageSignature]:
        if len(msghash) != 32:
            return None

        if not (signature := self._sign_raw_hash(msghash)):
            return None

        msg_sig = MessageSignature(**_convert_der_to_rsv(signature, 27))
        # TODO: Figure out how to properly compute v
        if not self.check_signature(msghash, msg_sig):
            msg_sig = MessageSignature(v=msg_sig.v + 1, r=msg_sig.r, s=msg_sig.s)

        return msg_sig

    def sign_message(self, msg: Any, **signer_options) -> Optional[MessageSignature]:
        if isinstance(msg, SignableMessage):
            message = msg
        if isinstance(msg, str):
            if msg.startswith("0x"):
                message = encode_defunct(hexstr=msg)
            else:
                message = encode_defunct(text=msg)
        if isinstance(msg, bytes):
            message = encode_defunct(primitive=msg)
        return self.sign_raw_msghash(_hash_eip191_message(message))

    def sign_transaction(self, txn: TransactionAPI, **signer_options) -> Optional[TransactionAPI]:
        """
        Sign an EIP-155 transaction.
        CHECK TYPE 0 transactions.

        Args:
            txn (``TransactionAPI``): A pydantic model of transaction data.

        Returns:
            TransactionAPI | None
        """
        unsigned_txn = serializable_unsigned_transaction_from_dict(txn.model_dump()).hash()
        if not (msg_sig := self._sign_raw_hash(unsigned_txn)):
            return None
        txn.signature = TransactionSignature(
            **_convert_der_to_rsv(msg_sig, (2 * txn.chain_id + 35) if txn.chain_id else 27)
        )
        # TODO: Figure out how to properly compute v
        if not self.check_signature(txn):
            txn.signature = TransactionSignature(
                v=txn.signature.v + 1,
                r=txn.signature.r,
                s=txn.signature.s,
            )

        return txn



import time
from typing import Optional

from ape.exceptions import ProviderError


class ApeAwsKmsProviderError(ProviderError):
    """
    An error related to the Hardhat network provider plugin.
    """


class ApeAwsKmsSubprocessError(ApeAwsKmsProviderError):
    """
    An error related to launching subprocesses to run Hardhat.
    """


class ApeAwsKmsTimeoutError(ApeAwsKmsSubprocessError):
    """
    A context-manager exception that raises if its operations exceed
    the given timeout seconds.
    This implementation was inspired from py-geth.
    """

    def __init__(
        self,
        message: Optional[str] = None,
        seconds: Optional[int] = None,
        exception: Optional[Exception] = None,
        *args,
        **kwargs,
    ):
        self._message = message or "Timed out waiting for process."
        self._seconds = seconds
        self._exception = exception
        self._start_time = None
        self._is_running = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def __str__(self):
        if self._seconds in [None, ""]:
            return ""

        return self._message

    @property
    def expire_at(self) -> float:
        if self._seconds is None:
            raise ValueError("Timeouts with 'seconds == None' do not have an expiration time.")
        elif self._start_time is None:
            raise ValueError("Timeout has not been started.")

        return self._start_time + self._seconds

    def start(self):
        if self._is_running is not None:
            raise ValueError("Timeout has already been started.")

        self._start_time = time.time()
        self._is_running = True

    def check(self):
        if self._is_running is None:
            raise ValueError("Timeout has not been started.")

        elif self._is_running is False:
            raise ValueError("Timeout has already been cancelled.")

        elif self._seconds is None:
            return

        elif time.time() > self.expire_at:
            self._is_running = False

            if isinstance(self._exception, type):
                raise self._exception(str(self))

            elif isinstance(self._exception, Exception):
                raise self._exception

            raise self

    def cancel(self):
        self._is_running = False


class RPCTimeoutError(ApeAwsKmsTimeoutError):
    def __init__(self, seconds=None, exception=None, *args, **kwargs):
        error_message = (
            "Timed out waiting for successful RPC connection to "
            f"the Hardhat node ({seconds} seconds) "
            "Try 'npx hardhat node' to debug."
        )
        super().__init__(
            message=error_message, seconds=seconds, exception=exception, *args, **kwargs
        )



from datetime import datetime
from typing import ClassVar

import boto3  # type: ignore[import]
from pydantic import BaseModel, Field


class AliasResponse(BaseModel):
    alias: str = Field(alias="AliasName")
    arn: str = Field(alias="AliasArn")
    key_id: str = Field(alias="TargetKeyId")
    creation: datetime = Field(alias="CreationDate")
    last_updated: datetime = Field(alias="LastUpdatedDate")


class KeyBaseModel(BaseModel):
    alias: str


class CreateKeyModel(KeyBaseModel):
    description: str = Field(alias="Description")
    policy: str | None = Field(default=None, alias="Policy")
    key_usage: str = Field(default="SIGN_VERIFY", alias="KeyUsage")
    key_spec: str = Field(default="ECC_SECG_P256K1", alias="KeySpec")
    admins: list[str] = []
    users: list[str] = []
    tags: list[dict[str, str]] | None = Field(default=None, alias="Tags")
    multi_region: bool | None = Field(default=None, alias="MultiRegion")
    origin: str = Field(alias="Origin")
    ADMIN_KEY_POLICY: ClassVar[
        str
    ] = """{
        "Version": "2012-10-17",
        "Id": "key-default-1",
        "Statement": [{
            "Sid": "Enable IAM User Permissions",
            "Effect": "Allow",
            "Principal": {"AWS": "{arn}"},
            "Action": "kms:*",
            "Resource": "*"
        }]
    }"""
    USER_KEY_POLICY: ClassVar[
        str
    ] = """{
        "Version": "2012-10-17",
        "Id": "key-default-1",
        "Statement": [{
            "Sid": "Allow use of the key",
            "Effect": "Allow",
            "Principal": {"AWS": "{arn}"},
            "Action": ["kms:Sign", "kms:Verify"],
            "Resource": "*"
        }]
    }"""

    def to_aws_dict(self):
        alias_dict = {}
        for k, v in self.model_dump().items():
            field = self.model_fields[k]
            if field.alias and v:
                alias_dict[field.alias] = v
        return alias_dict


class CreateKey(CreateKeyModel):
    origin: str = Field(default="AWS_KMS", alias="Origin")


class ImportKey(CreateKeyModel):
    origin: str = Field(default="EXTERNAL", alias="Origin")


class DeleteKey(KeyBaseModel):
    key_id: str
    days: int = 30


class KmsClient:
    def __init__(self):
        self.client = boto3.client("kms")

    @property
    def raw_aliases(self) -> list[AliasResponse]:
        paginator = self.client.get_paginator("list_aliases")
        pages = paginator.paginate()
        return [
            AliasResponse(**page)
            for alias_data in pages
            for page in alias_data["Aliases"]
            if "alias/aws" not in page["AliasName"]
        ]

    def get_public_key(self, key_id: str):
        return self.client.get_public_key(KeyId=key_id)["PublicKey"]

    def sign(self, key_id, msghash):
        response = self.client.sign(
            KeyId=key_id,
            Message=msghash,
            MessageType="DIGEST",
            SigningAlgorithm="ECDSA_SHA_256",
        )
        return response.get("Signature")

    def create_key(self, key_spec: CreateKey):
        response = self.client.create_key(**key_spec.to_aws_dict())
        key_id = response["KeyMetadata"]["KeyId"]
        self.client.create_alias(
            AliasName=f"alias/{key_spec.alias}",
            TargetKeyId=key_id,
        )
        if key_spec.tags:
            self.client.tag_resource(
                KeyId=key_id,
                Tags=key_spec.tags,
            )
        if key_spec.admins:
            for arn in key_spec.admins:
                self.client.put_key_policy(
                    KeyId=key_id,
                    PolicyName="default",
                    Policy=key_spec.ADMIN_KEY_POLICY.format(arn=arn),
                )
        if key_spec.users:
            for arn in key_spec.users:
                kms_client.client.put_key_policy(
                    KeyId=key_id,
                    PolicyName="default",
                    Policy=key_spec.USER_KEY_POLICY.format(arn=arn),
                )
        return key_id

    def delete_key(self, key_spec: DeleteKey):
        self.client.delete_alias(AliasName=key_spec.alias)
        self.client.schedule_key_deletion(KeyId=key_spec.key_id, PendingWindowInDays=key_spec.days)
        return key_spec.alias


class IamClient:
    def __init__(self):
        self.client = boto3.client("iam")

    def list_users(self):
        result = self.client.list_users()
        return result.get("Users")

    def list_admins(self):
        admins = []
        for user in self.list_users():
            user_name = user["UserName"]
            user_policies = self.client.list_attached_user_policies(UserName=user_name)
            for policy in user_policies["AttachedPolicies"]:
                if policy["PolicyName"] == "AdministratorAccess":
                    admins.append(user_name)
        return admins


kms_client = KmsClient()
iam_client = IamClient()



import click
from ape.cli import ape_cli_context

from ape_aws.client import CreateKey, DeleteKey, kms_client


@click.group("kms")
def kms():
    """Manage AWS KMS keys"""


@kms.command(name="create")
@ape_cli_context()
@click.option(
    "-a",
    "--admin",
    "administrators",
    multiple=True,
    help="Apply key policy to a list of administrators if applicable, ex. -a ARN1, -a ARN2",
    metavar="list[ARN]",
)
@click.option(
    "-u",
    "--user",
    "users",
    multiple=True,
    help="Apply key policy to a list of users if applicable, ex. -u ARN1, -u ARN2",
    metavar="list[ARN]",
)
@click.argument("alias_name")
@click.argument("description")
def create_key(
    cli_ctx,
    alias_name: str,
    description: str,
    administrators: list[str],
    users: list[str],
):
    """
    Create an Ethereum Private Key in AWS KmsAccount

    \b
    Args:
        alias_name str: The alias of the key you intend to create
        description str: The description of the key you intend to create.
    """
    key_spec = CreateKey(
        alias=alias_name,
        Description=description,
        admins=administrators,
        users=users,
    )
    key_id = kms_client.create_key(key_spec)
    cli_ctx.logger.success(f"Key created successfully with ID: {key_id}")


# TODO: Add `ape aws kms import`
# TODO: Add `ape aws kms sign-message [message]`
# TODO: Add `ape aws kms verify-message [message] [hex-signature]`


@kms.command(name="delete")
@ape_cli_context()
@click.argument("alias_name")
@click.option("-d", "--days", default=30, help="Number of days until key is deactivated")
def schedule_delete_key(cli_ctx, alias_name, days):
    if "alias" not in alias_name:
        alias_name = f"alias/{alias_name}"
    kms_account = None
    for account in kms_client.raw_aliases:
        if account.alias == alias_name:
            kms_account = account

    if not kms_account:
        cli_ctx.abort(f"No KMS Key with alias name: {alias_name}")

    delete_key_spec = DeleteKey(alias=alias_name, key_id=kms_account.key_id, days=days)
    key_alias = kms_client.delete_key(delete_key_spec)
    cli_ctx.logger.success(f"Key {key_alias} scheduled for deletion in {days} days")



import click

from ape_aws.client import iam_client


@click.group("iam")
def iam():
    """Manage AWS Admin information"""


@iam.command()
def list_admins():
    if admins := "\n".join(iam_client.list_admins()):
        click.echo(f"Administrators:\n{admins}")


@iam.command()
def list_users():
    if users := "\n".join(iam_client.list_users()):
        click.echo(f"Users:\n{users}")
